import numpy as np

# from medil.simulate import MicroModel as mm

# from medil.grues import InputData as data

# mic = mm(5, 0.5)
# samp = mic.sample(1000)
# grues = data(samp)
# grues.mcmc(prior=(2, 3))

# grues.optimal_uec
# mic.uec
